/*!
 * \file
 * \brief   Main application file for PDS HW2 (MPI)
 *
 * \author
 *    Christos Choutouridis AEM:8997
 *    <cchoutou@ece.auth.gr>
 */

#include <exception>
#include <iostream>
#include <algorithm>  // rand/srand
//#include <ctime>      // rand/srand
#if !defined TESTING
  #include <mpi.h>
#endif

#include "distbitonic.hpp"
#include "utils.hpp"
#include "config.h"
//#include "matrix.hpp"

// Global session data
session_t  session;
mpi_t      mpi;

/*
 * Sorting data for up to 8 processes
 */
AllData_t Data {
        Data_t (8),
        Data_t (8),
        Data_t (8),
        Data_t (8),
        Data_t (8),
        Data_t (8),
        Data_t (8),
        Data_t (8)
};


/*!
 * A small command line argument parser
 * \return  The status of the operation
 */
bool get_options(int argc, char* argv[]){
    bool status =true;

    // iterate over the passed arguments
    for (int i=1 ; i<argc ; ++i) {
        std::string arg(argv[i]);     // get current argument

        if (arg == "-x" || arg == "--xxxxx") {
            if (i+2 < argc) {
//                session.corpusMtxFile = std::string(argv[++i]);
//                session.corpusDataSet = std::string(argv[++i]);
            }
            else
                status = false;
        }

        else if (arg == "-v" || arg == "--verbose")
            session.verbose = true;
        else if (arg == "-h" || arg == "--help") {
            std::cout << "distbitonic - A distributed bitonic sort\n\n";
            std::cout << "distbitonic -x <> [-v]\n";
            std::cout << '\n';
            std::cout << "Options:\n\n";
            std::cout << "   -v | --verbose\n";
            std::cout << "      Request a more verbose output to stdout.\n\n";
            std::cout << "   -h | --help\n";
            std::cout << "      Prints this and exit.\n\n";
            std::cout << "Examples:\n\n";
            std::cout << "   ...Example case...:\n";
            std::cout << "   > distbitonic -x <xxxxx>  \n\n";

            exit(0);
        }
        else {   // parse error
            std::cout << "Invocation error. Try -h for details.\n";
            status = false;
        }
    }

    return status;
}



#if !defined TESTING
int main(int argc, char* argv[]) try {
    // try to read command line
    if (!get_options(argc, argv))
        exit(1);

    // Initialize the MPI environment
    MPI_Init(NULL, NULL);

    #if defined DEBUG
        /*
         * In case of a debug build we will wait here until sleep_wait
         * will reset via debugger. In order to do that the user must attach
         * debugger to all processes. For example:
         *  $> mpirun -np 2 ./<program path>
         *  $> ps aux | grep <program>
         *  $> gdb <program> <PID1>
         *  $> gdb <program> <PID2>
         */
    #if defined TESTING
        volatile bool sleep_wait = false;
    #else
        volatile bool sleep_wait = true;
    #endif
        while (sleep_wait)
            sleep(1);
    #endif

    // Get the number of processes
    MPI_Comm_size(MPI_COMM_WORLD, reinterpret_cast<int *>(&mpi.world_size));

    // Get the rank of the process
    MPI_Comm_rank(MPI_COMM_WORLD, reinterpret_cast<int *>(&mpi.world_rank));

    // Get the name of the processor
    char processor_name[MPI_MAX_PROCESSOR_NAME];
    int name_len;
    MPI_Get_processor_name(processor_name, &name_len);
    mpi.processor_name = std::string (processor_name, name_len);

    // Print off a hello world message
    std::cout   << "Hello world from processor: " << mpi.processor_name
                << " rank " << mpi.world_rank
                << " out of " << mpi.world_size << " processors\n";

//    std::srand(unsigned(std::time(nullptr)));
//    for (auto& v : Data) {
//        std::generate(v.begin(), v.end(), std::rand);
//    }
//
//    distbitonic (2, Data);
//    distbitonic (4, Data);

    // Finalize the MPI environment.
    MPI_Finalize();
    return 0;
}
catch (std::exception& e) {
    //we probably pollute the user's screen. Comment `cerr << ...` if you don't like it.
    std::cerr << "Error: " << e.what() << '\n';
    exit(1);
}

#else

#include <gtest/gtest.h>
#include <exception>

GTEST_API_ int main(int argc, char **argv) try {
   testing::InitGoogleTest(&argc, argv);
   return RUN_ALL_TESTS();
}
catch (std::exception& e) {
    std::cout << "Exception: " << e.what() << '\n';
}

#endif